#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <math.h>
#include <signal.h>
#include "common.h"
#include "libtest.h"

bool signl = false, contin = true, limits = true;

static void handler( int signo ) {                      // подведение результатов теста
   contin = false;
}

int main( int argc, char *argv[] ) {
   int opt, rep = 1000, freq = RTC_MAX_FREQ;            // 8192 hz 
   while ( ( opt = getopt( argc, argv, "f:r:lsv") ) != -1 ) {
      switch( opt ) {
         case 'f' : if( atoi( optarg ) > 0 ) freq = atoi( optarg ); break;
         case 'r' : if( atoi( optarg ) > 0 ) rep = atoi( optarg ); break;
         case 'l' : limits = false; break;
         case 's' : signal( SIGINT, handler ); signl = true; break;
         case 'v' : debug_level++; break;
         default :
            fprintf( stderr, "запуск: %s [-f <2**n>] [-r <n>]\n", argv[ 0 ] );
            exit( EXIT_FAILURE );
      }
   }  
   if( argc !=  optind)
      err_dump( "параметры команды не допускаются!" );
   int fd = open( "/dev/rtc", O_RDONLY );               // открыть RTC
   if( fd < 0 )
      err_dump( "у вас должны быть права root!" ); 
   if( freq > RTC_MAX_FREQ )
      err_dump( "максимальная частота прерываний RTC 8192 hz!" );   
   set_rt( 75 );
   const uint64_t nper = 1000000000UL / freq, limb = nper >> 1, limt = nper << 1; 
   fprintf( stdout, "период RTC заказан %.2f us [%lu ns]\n",
            1000000. / freq, nper );
   if( limits )
      fprintf( stdout, "допустимые пределы: %lu...%lu ns\n", limb, limt ); 
   mlockall( MCL_CURRENT );                             // запретить свопирование страниц
   fprintf( stdout, "инициализация..." );
   fflush( stdout );

   uint64_t cali = calibr( 1000 );
   uint64_t hz = proc_hz();

   if( debug_level ) 
      fprintf( stdout, "\rтактовая частота процессора = %lu\n", (long)hz );
   else
      fprintf( stdout, "\r""накопление...   " );
   ioctl( fd, RTC_PIE_ON, 0 );                          // разрешить периодические прерывания
   double s1 = 0., s2 = 0.;
   unsigned long data;
   ioctl( fd, RTC_IRQP_SET, freq );                     // установка частоты RTC
   read( fd, &data, sizeof(unsigned long) );            // начальная синхронизация
   uint64_t min = 0, max = 0, vin = 0, vax = 0; 
   uint64_t n = 0;
   for( uint64_t i = 0; i < rep; i += signl ? 0 : 1 ) { // выполнять периодическую работу ...
      uint64_t next = rdtsc() + cali;
      read( fd, &data, sizeof(unsigned long) );         // блокировать до следующего периодического прерывния
      uint64_t t = rdtsc() - next; 
      if( !contin ) break;                              // завершение по сигналу
      if( 0 == n ) min = t;
      if( debug_level ) {
         if( !signl && rep <= 100 ) 
            fprintf( stdout, "%ld[%.2f] ", t, tick2us( hz, t ) );
         else if( t > max || t < min )
            fprintf( stdout, "%ld%c:%ld[%.2f] ", 
                     n, ( t > max ? '+' : '-' ), t, tick2us( hz, t ) );
         fflush( stdout );
      }
      if( limits ) {                                    // выбросы? - не понимаю!!!
         if( t < limb ) {
            vin++;
            continue;
         }
         if( t > limt ) {
            vax++;
            continue;
         }
      }
      n++;
      max = t > max ? t : max;
      min = t < min ? t : min;
      s1 += (double)t;                                  // накапливать латентность
      s2 += (double)t * t;
   }
   ioctl( fd, RTC_PIE_OFF, 0 );                         // запретить периодические прерывания
   close( fd );
   fprintf( stdout, "\n" );
   usleep( 200 );
   s1 /= n;
   s2 = sqrt( s2 / n - s1 * s1 );
   fprintf( stdout, "\r" );
   if( debug_level > 1 ) 
      fprintf( stdout, "n=%lu | mean=%f | sko=%f | min=%lu | max=%lu \n",
               n, s1, s2, min, max );
   if( ( vin || vax ) && limits )
      fprintf( stdout, "число выбросов: вниз = %lu | вверх = %lu \n",
               vin, vax );
   fprintf( stdout, "n=%lu : %.2f/%.2f/%.2f us [+/-%.1f%c]\n",
            (long)n, tick2us( hz, min ), 
            tick2us( hz, (unsigned long)round( s1 ) ),
            tick2us( hz, max ), 
            s2 / s1 * 100, '%' );
   exit( EXIT_SUCCESS );
}
