#include <signal.h>
#include <time.h>
#include <math.h>
#include <errno.h>
#include "common.h"
#include "libtest.h" 

unsigned mksleep = 100000;                        // период мксек.
bool contin = true, noserver; 
uint64_t count = 0, cali, hz;
double s1 = 0., s2 = 0.;

static void handler( int signo ) {                // подведение результатов теста
   contin = false;
   fprintf( stdout, "\rрезультат счёта:\n" );
   fprintf( stdout, "число пульсов = %llu\n", (long long)count );
   s1 /= count;
   fprintf( stdout, "средняя задержка: %.0f тактов | %.2f мкс.\n", 
            round( s1 ), tick2us( hz, (unsigned long)round( s1 ) ) );
   s2 = sqrt( s2 / count - s1 * s1 );
   fprintf( stdout, "с СКО отклонением = %.2f (%.1f%c)\n", s2, s2 / s1 * 100, '%' );  
   exit( 0 );
}

static void halarm( int signo ) {                 // тайм-аут отсутсвия сервера
   err_dump( "dg_cli: the server is not responding" );
}

void dg_cli( int sockfd,
             struct sockaddr* pserv_addr,         // ptr to appropriate sockaddr_XX structure 
             int servlen ) {	                  // actual sizeof(*pserv_addr) 
   int n;
   dgram_t sendline, recvline;
   while( true ) {
      n = MAXLINE;
      signal( SIGALRM, halarm );                  // установить тайм-аут
      alarm( 3 );
      sendline.clc = rdtsc();
      if( sendto( sockfd, (void*)&sendline, n, 0, pserv_addr, servlen ) != n )
         err_dump( "dg_cli: sendto error on socket" );
      n = recvfrom( sockfd, (void*)&recvline, MAXLINE, 0, NULL, NULL );
      double ns = (double)( rdtsc() - recvline.clc - cali );
      signal( SIGALRM, SIG_IGN );                 // снять тайм-аут
      if( n < 0 )
         err_dump( "dg_cli: recvfrom error on socket" );
      if( debug_level > 2 )
         fprintf( stdout, "%.0f ticks\n", round( ns ) );
      s1 += ns;
      s2 += ns * ns;
      count++;
      usleep( mksleep ); 
   }
}


int main( int argc, char* argv[] ) {
   int c;
   const char sopt[] = "t:s:vh";
   char host_addr[ 120 ] = "127.0.0.1";      
   while( -1 != ( c = getopt( argc, argv, sopt ) ) )
      switch( c ) {
         case 't':
            mksleep = atoi( optarg );
            break;
         case 's':
            strncpy( host_addr, optarg, sizeof( host_addr ) ); 
            break;
         case 'v':
            debug_level++;
            break;
         case 'h':
            fprintf( stderr, "%s [-t <микросекунд>][-s <сервер>][-vv...]\n", argv[ 0 ] );
            return( EXIT_SUCCESS );
         default:
            fprintf( stderr, "допустимые опции: " );
            for( c = 0; c < strlen( sopt ); ) {
               if( ':' == sopt[ c ] ) 
                  fprintf( stderr, "<...>" );
               else 
                  fprintf( stderr, "-%c", sopt[ c ] );
               c++;
               if( c < strlen( sopt ) ) {
                 if( ':' != sopt[ c ] )
                    fprintf( stderr, ", " );
               }
               else         
                  fprintf( stderr, "\n" ); 
            }         
            return( EXIT_FAILURE );
      }
   if( debug_level )
      fprintf( stdout, "server: %s, interval=%u мкс.\n", host_addr, mksleep );  
   hz = proc_hz();
   if( debug_level > 1 )
      fprintf( stdout, "тактовая частота процессора: %llu Hz\n", (long long)hz );  
   fprintf( stdout, "для нормального завершения: Ctrl+C!\n" );
   cali = calibr( 1000 );
   signal( SIGINT, handler );
   int sockfd;                      // Open a UDP socket (an Internet datagram socket).
   if( ( sockfd = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 )
      err_dump( "client: can't open datagram socket" );
   struct sockaddr_in serv_addr;    // Fill in the structure "serv_addr" with the address of the server
   bzero( (char*)&serv_addr, sizeof( serv_addr ) );
   serv_addr.sin_family      = AF_INET;
   serv_addr.sin_addr.s_addr = inet_addr( host_addr );
   serv_addr.sin_port        = htons( SERV_UDP_PORT );
   dg_cli( sockfd, (struct sockaddr*)&serv_addr, sizeof( serv_addr ) );   
   close( sockfd );
   exit( 0 );
}
